#! /usr/bin/ksh
# ftplib.ksh - ksh FTP library
# Heiner Steven (heiner.steven@odn.de) 2000
# @(#)ftplib.ksh	1.1
#
# This library directly implements a limited set of the
# File Transfer Protocol (FTP) (RFC 959)
#
# Notes:
#    o	uses "telnet" for the control connection, and
#	"netcat" for each data connection
#
# user pi  = user protocol interpreter (control connection)
# user dtp = user data transfer process (data connection)
#
# TODO:
#    o	set traps to terminate processes
#
# Notes
#    o	The current usage of "netcat" has some drawbacks:
#	 -  STOR: netcat will not terminate automatically after
#	    transferring all data
#	 -  At this time we do not know, if the netcat process is
#	    already running. This causes timing problems.
#	 -  There should be a way to determine the listen port number
#	    "netcat" uses (instead of using fixed port numbers)

PN=ftplib

#########################################################################
# Global variables
#########################################################################

ftp_response=

# FTP reply codes
ftp_code=000				# three digit code, i.e. 200
ftp_code1=0; ftp_code2=0; ftp_code3=0	# first, second, third digit

#########################################################################
# Local variables
#########################################################################

_ftp_verbose=1
_ftp_passive=0
_ftp_debug=2

_ftp_listenport=8901

_ftp_pi_pid=0		# "Protocol Interpreter" PID
_ftp_dtp_pid=0		# "Data Transfer Process" PID

# We use the convention that IP addresss variables named "...addr"
# contain a numeric IP address, while "...host" variables can
# contain a resolvable host name or a numeric IP address
_ftp_localaddr=		# Local IP address (numeric)
_ftp_localhost=

_ftp_kshflags=${-##!(*[xvu]*)}
#print -u2 "\$-=<$->, kshflags=<$_ftp_kshflags>"

#########################################################################

function _ftp_error {
    print -u2 "$PN: $@"
}

#########################################################################

function _ftp_fatal {
    _ftp_error "$@"
    exit 3
}

#########################################################################
# ftplib - initializing function
#########################################################################

function ftplib
{
    _ftp_localhost=$(uname -n)
    # Is our ip address in /etc/hosts?
    _ftp_localaddr=$(grep "$_ftp_localhost" /etc/hosts)
    _ftp_localaddr=${_ftp_localaddr%%@([ 	])*}
    if [[ -z "$_ftp_localaddr" ]]
    then			# using DNS?
	_ftp_localaddr=$(nslookup "$_ftp_localhost" | grep Address:)
	_ftp_localaddr=${_ftp_localaddr##Address:*([ 	])}
    fi
    [[ -n "$_ftp_localaddr" ]] ||
    	_ftp_fatal "$0: cannot determine own ip address for $_ftp_localhost"
}

#########################################################################
# _ftp_toftpstr - convert number to comma-separated list of bytes
#
# Example: "8901" -> "34,197"
#########################################################################

function _ftp_toftpstr
{
    print $(print "for (v=$1;v;v=v/256) b[i++]=v%256; while (--i>=0) b[i]" |
    	bc) | tr ' ' ,
}

function _ftp_fromftpstr
{
    integer var result=0

    for var in $(echo "$1" | tr , ' ')
    do result=result*256+var
    done
    print $result
}

#########################################################################
# _ftp_gethostbyname
#########################################################################

function _ftp_gethostbyname # host ipaddr
{
    (( $# == 2 )) || _ftp_fatal "$0: host [ipaddr]"

    typeset host=$1
    typeset vn=$2
    typeset ip
    
    # First lookup /etc/hosts

    ip=$(grep "[ 	][ 	]*$host[ 	][ 	]*" < /etc/hosts ||
    	grep "[ 	][ 	]*$host\$" < /etc/hosts)
    if [[ -n $ip ]]
    then
    	ip=${ip%%@([ 	])*}
    else
    	# Then use DNS, if configured
	ip=$(nslookup "$host" | grep Address:)
	if [[ -n $ip ]]
	then
	    ip=${ip##Address:*([ 	])}
	fi
    fi
    eval $vn='"$ip"'
}

#########################################################################

function _ftp_sendline {
    (( _ftp_debug >= 2 )) && print -u2 "$PN: FTP: > \"$@\""
    print -u9 -- "$@"
}

#########################################################################

function _ftp_readline {
    (( $# < 1 )) && _ftp_fatal "$0: need at least 1 argument"
    typeset vn=$1			# variable name
    #typeset OIFS=$IFS IFS="
#"
    if read -u8 -r
    then
    	#IFS=$OIFS
	(( _ftp_debug >= 3 )) && print -u2 "$PN: FTP: < \"$REPLY\""
	eval $vn='"$REPLY"'
	return 0
    fi
    return 1
}

#########################################################################

function _ftp_getack {
    typeset ackline
    typeset -L4 respcode
    typeset -L3 firstcode
    typeset -L3 code

    _ftp_readline ackline || return $?
    respcode=$ackline		# first four characters
    code=$ackline		# first three characters
    firstcode=$code

    # If the character following the response code is a hyphen
    # ("-"), a multi-line response follows.  The last line has the
    # same response code than the first line, but with a space
    # character following the response code, i.e.
    #	200-login message
    #	    this is a message
    #	200 please login

    until [[ $respcode = "$firstcode " ]]
    do
	_ftp_readline ackline || break
	respcode=$ackline
	code=$respcode
    done

    ftp_response=${ackline#????}
    ftp_code=$code
    ftp_code1=${code%??}
    ftp_code2=${ftp_code#?}
    ftp_code3=${ftp_code#??}

    case "$code" in
	[123]*)	return 0;;
	4*)	return 1;;
	5*)	return 2;;
	*)	print -u2 "NOTE: unexpected response: $ackline";;
    esac
    return 3
}

#########################################################################
# Helper functions
#########################################################################

function _ftp_shutdown_dtp
{
    if (( _ftp_dtp_pid >= 0 ))
    then
    	_ftp_terminate_process _ftp_dtp_pid
	return 0
    else
    	return 1
    fi
}

#########################################################################
# _ftp_startup_listen_dtp [-l listenportvname] [-p pidvname]
#
# Starts a listening process, and returns the listen port numer and
# the process id of the listening process. If a listen port number was
# specified, it is used, otherwise a random listen port will be
# chosen.
#########################################################################

function _ftp_startup_listen_dtp # -p portvname
{
    typeset portvname port
    typeset pidvname pid
    typeset opt options
    typeset pipe=${TMPDIR:=/tmp}/ftplib.$$.p

    while getopts :l:o:p: opt
    do
    	case "$opt" in
	    l)	portvname=$OPTARG
	    	eval port=\"'$'$OPTARG\";;
	    o)	options=$OPTARG;;
	    p)	pidvname=$OPTARG;;
	    *)	_ftp_error "$0: invalid argument: -$opt"; return 1;;
	esac
    done

    # "netcat" will write the chosen listen port number to stderr.
    # We will redirect the output to a named pipe that we read
    # later on to determine the port number.

    mknod "$pipe" p ||
    	    { _ftp_error "$0: could not create named pipe"; return 1; }

    if [[ -n $port ]]
    then netcat -v -l -p "$port" 2> "$pipe" & pid=$!
    else netcat -v -l		 2> "$pipe" & pid=$!
    fi

    [[ -n $pidvname ]] && eval $pidvname=\"$pid\"

    # The first line of the "netcat" output contains the port number.

    read firstline < "$pipe" ||
    	    { _ftp_error "$0: could not read named pipe"; return 1; }

    # "netcat" would block on further output to stderr, so we will
    # empty the named pipe, and throw the data away.
    # Since "cat" keeps the pipe open, we can remove it immediately.
    # The entry will the be removed after "cat" terminates.

    < "$pipe" cat & #> /dev/null 2>&1 &
    rm -f "$pipe" # >/dev/null 2>&1

    # Example output: "listening on [any] 2656 ..."

    if [[ $firstline = *listening*on* ]]
    then
    	port=$(print "$firstline" | cut -d' ' -f4)
	[[ -n $portvname ]] && eval $portvname=\"$port\"
	return 0
    else
    	_ftp_error "$0: unexpected response: \"$firstline\""
	return 1
    fi
}

#########################################################################
# _ftp_startup_dtp [-p pidvname] host port
#
# Starts a data transfer process, and returns the the process id of
# process.
#########################################################################

function _ftp_startup_dtp # [-p pidvname] host port
{
    typeset host port
    typeset pidvname pid
    typeset opt options
    typeset pipe=${TMPDIR:=/tmp}/ftplib.$$.p

    while getopts :l:o:p: opt
    do
    	case "$opt" in
	    p)	pidvname=$OPTARG;;
	    o)	options=$OPTARG;;
	    *)	_ftp_error "$0: invalid argument: -$opt"; return 1;;
	esac
    done
    shift OPTIND-1

    (( $# < 2 )) &&
    	    { _ftp_error "$0: need host name and port number"; return 1; }

    host=$1
    port=$2

    # We will only return to the caller, if the "netcat" process is
    # ready to transfer data. Therefore we redirect the output to a
    # named pipe, and wait for some data to appear...

    mknod "$pipe" p ||
    	    { _ftp_error "$0: could not create named pipe"; return 1; }

    netcat -v $options "$host" "$port" 2> "$pipe" & pid=$!

    [[ -n $pidvname ]] && eval $pidvname=\"$pid\"

    # The following read will block, until netcat prints it's first
    # line of output to stderr:

    read firstline < "$pipe" ||
    	    { _ftp_error "$0: could not read named pipe"; return 1; }

    # "netcat" would block on further output to stderr, so we will
    # empty the named pipe, and throw the data away.
    # Since "cat" keeps the pipe open, we can remove it immediately.
    # The entry will the be removed after "cat" terminates.

    < "$pipe" cat & #> /dev/null 2>&1 &
    rm -f "$pipe" # >/dev/null 2>&1

    return 0
}

#########################################################################
# _ftp_setupconn_active - setup data connection ("active") mode
#
# Prepares the server for a subsequent data connection handled by
# the dtp ("data transfer process").
#
# Note that "active" means: the remote server actively connects to
# our dtp, we have to listen for an incoming request.
#
# Global variables:
#	_ftp_localaddr		(R)
#	_ftp_listenport		(R/W)
#
# Returns
#	0			no error
#	>=1			error
#########################################################################

function _ftp_setupconn_active
{
    (( _ftp_debug > 0 )) && print -u2 "DEBUG: $0 $@"
    (( _ftp_debug > 3 )) && set -$_ftp_kshflags

    typeset localaddr localaddrvar
    typeset localport localportvar
    typeset ipstr opt

    while getopts :a:p: opt
    do
    	case "$opt" in
	    a)	localaddrvar=$OPTARG
	    	eval localaddr='$'$OPTARG;;
	    p)	localportvar=$OPTARG
	    	eval localport='$'$OPTARG;;
	    *)	_ftp_fatal "$0: invalid argument: -$opt";;
	esac
    done

    if [[ -z $localport ]]
    then
    	localport=$_ftp_listenport
	((_ftp_listenport=_ftp_listenport+1))
	[[ -n $localportvar ]] && eval $localportvar=\"$localport\"
    fi

    if [[ -z $localaddr ]]
    then
    	localaddr=$_ftp_localaddr
	[[ -n $localaddrvar ]] && eval $localaddrvar='"$localaddr"'
    fi

    ipstr=$(print "$localaddr" | tr . ,),$(_ftp_toftpstr "$localport") 
    if _ftp_sendline PORT "$ipstr" && _ftp_getack
    then return 0
    else return 1
    fi
}

#########################################################################
# _ftp_setupconn_passive - setup data connection ("passive") mode
#
# Prepares the server for a subsequent data connection handled by
# the dtp ("data transfer process").
#
# Note that "passive" means: the remote server passively waits for
# our dtp to connect.
#
# Global variables:
#	ftp_response		(R)
#
# Returns
#	0			no error
#	>=1			error
#########################################################################

function _ftp_setupconn_passive
{
    (( _ftp_debug > 0 )) && print -u2 "DEBUG: $0 $@"
    (( _ftp_debug > 3 )) && set -$_ftp_kshflags

    typeset remoteaddr remoteaddrvar 
    typeset remoteport remoteportvar
    typeset opt ipstring

    while getopts :a:p: opt
    do
    	case "$opt" in
	    a)	remoteaddrvar=$OPTARG;;
	    p)	remoteportvar=$OPTARG;;
	    *)	_ftp_fatal "$0: invalid argument: -$opt";;
	esac
    done

    if _ftp_sendline PASV && _ftp_getack
    then
	# Extract the server's IP address and port number from the
	# response. Example:
	#   "227 Entering Passive Mode (127,0,0,1,5,12)"
	ipstr=$(expr "$ftp_response" : '.*[^0-9]\([0-9][0-9]*,[0-9][0-9]*,[0-9][0-9]*,[0-9][0-9]*,[0-9][0-9]*,[0-9][0-9]*\).*')
	remoteaddr=$(echo "$ipstr" | cut -d, -f1-4 | tr , .)
	remoteport=$(_ftp_fromftpstr $(echo "$ipstr" | cut -d, -f5-6) )
	#print -u2 "DEBUG: ip <$remoteaddr> port=<$remoteport>"

	[[ -n $remoteaddrvar ]] && eval $remoteaddrvar='"$remoteaddr"'
	[[ -n $remoteportvar ]] && eval $remoteportvar='"$remoteport"'
	return 0
    else
    	return 1
    fi
}

#########################################################################
# _ftp_setup_dtp - start "data transfer process"
#
# Data from stdin is written to the remote server-dtp ("data
# transfer process"). Data from the server-dtp is written to
# stdout.
#
# Global variables:
#	_ftp_passive		(R)
#########################################################################

function _ftp_setup_dtp
{
    (( _ftp_debug > 0  )) && print -u2 "DEBUG: $0 $@"
    (( _ftp_debug > 3 )) && set -$_ftp_kshflags

    typeset remaddr
    integer remport locport
    typeset opt options
    typeset dtppid pidvname

    while getopts :o:p: opt
    do
    	case "$opt" in
	    p)	pidvname=$OPTARG;;
	    o)	options=$OPTARG;;
	    *)	print -u2 "$0: invalid option -$opt"; exit 1;;
	esac
    done

    if (( _ftp_passive ))
    then
    	# Server waits for our data connection

	if _ftp_setupconn_passive -a remaddr -p remport
	then
	    [[ -n "$remaddr" ]] || _ftp_fatal "$0: need server's address"
	    [[ -n "$remport" ]] || _ftp_fatal "$0: need server's port"
	    _ftp_startup_dtp -p dtppid -o "-w 1" "$remaddr" "$remport"
	    [[ -n $pidvname ]] && eval $pidvname=\"$dtppid\"
	    return 0
	else
	    return 1
	fi
    else
	# Server will actively contact us; set up listening process We
	# will first start up the dtp, and then issue the PORT
	# command. Otherwise the server may try to contact us before
	# the dtp process is up and running.

	if _ftp_startup_listen_dtp -l locport -p dtppid -o "$options"
	then
	    [[ -n $pidvname ]] && eval $pidvname=\"$dtppid\"
	    #print -u2 "DEBUG: $0: pid=$dtppid, listenport=$locport"
	    _ftp_setupconn_active -p locport
	    return 0
	else
	    _ftp_error "$0: could not start data transfer process"
	    return 1
	fi
    fi
}

#########################################################################

function _ftp_terminate_process # pidname
{
    (( _ftp_debug > 0 )) && print -u2 "DEBUG: $0 $@"
    (( _ftp_debug > 3 )) && set -$_ftp_kshflags

    (( $# < 1 )) && _ftp_fatal "$0: specify pidvariable name"

    typeset pidvname=$1 pid
    eval pid=\"$pidvname\"

    # Give it time to cleanup

    if kill -0 $pid >/dev/null 2>&1
    then
	kill -TERM $pid ||
	(kill -HUP $pid; res=$?; sleep 1; exit $res) ||
	kill -KILL $pid
    fi >/dev/null 2>&1
    eval $pidvname=0
    return 0
}

#########################################################################

function _ftp_startup_pi
{
    (( _ftp_debug > 0 )) && print -u2 "DEBUG: $0 $@"
    (( _ftp_debug > 3 )) && set -$_ftp_kshflags

    if (( _ftp_pi_pid <= 0 ))
    then
	# Starting telnet co-process.
	# Typical output examples:
	# Success:
	#	Trying 127.0.0.1...
	#	Connected to localhost.
	#	Escape character is '^]'.
	# Invalid host name:
	#	gnuelpf: Unknown host
	# Host without FTP server:
	#	Trying 192.67.198.33...
	#	telnet: Unable to connect to remote host: Connection refused

	telnet "$@" 2>&1 |&
	_ftp_pi_pid=$!

	exec 8<&p 9>&p
	typeset response errorcode

	if  _ftp_readline response && [[ $response = Trying* ]] &&
	    _ftp_readline response && [[ $response = Connected* ]] &&
	    _ftp_readline response
	then
	    echo >&2 "$0: DEBUG: startup o.k."
	    _ftp_remotehost=$host
	    return 0
	else
	    errorcode=$?
	    _ftp_error "$0: could not startup control process: \"$response\""
	    return $errorcode
	fi
    else
    	return 1
    fi
}

#########################################################################

function _ftp_shutdown_pi
{
    (( _ftp_debug > 0 )) && print -u2 "DEBUG: $0 $@"
    (( _ftp_debug > 3 )) && set -$_ftp_kshflags

    if (( _ftp_pi_pid >= 0 ))
    then
	# Close stdin and stdout file descriptors of the protocol interpreter
	# process. This should terminate it.

	exec 8<&- 9<&-
	sleep 1
	if kill -0 $_ftp_pi_pid 2>/dev/null
	then				# co-process is still running
	    _ftp_terminate_process _ftp_pi_pid
	else
	    ((_ftp_pi_pid=0))
	fi
	return 0
    fi
    return 1
}

#########################################################################
#
# FTP commands
#
#########################################################################

#########################################################################

function ftp_open # host [port=21]
{
    (( _ftp_debug > 0  )) && print -u2 "DEBUG: $0 $@"
    (( _ftp_debug > 3 )) && set -$_ftp_kshflags
    (( $# < 1 )) && _ftp_fatal "$0: hostname [port]"

    typeset host=$1
    typeset port=${2:-21}
    typeset ipaddr

    ((_ftp_pi_pid > 0 )) && _ftp_shutdown_pi

    # Special handling: If we connect to the local host,
    # use the loop-back ip addresses for the data processing
    # process:

    _ftp_gethostbyname "$host" ipaddr
    [[ $ipaddr = 127* ]] && _ftp_localaddr=$ipaddr

    _ftp_startup_pi "$host" "$port" ||
    	_ftp_fatal "$ftp_error"

: '
    netcat "$host" "$port" 2>&1 |& _ftp_pi_id=$!
    exec 8<&p 9>&p
'
    _ftp_getack			# read welcome message
}

#########################################################################

function ftp_quit
{
    (( _ftp_debug > 0 )) && print -u2 "DEBUG: $0 $@"
    (( _ftp_debug > 3 )) && set -$_ftp_kshflags

    _ftp_sendline QUIT && _ftp_getack
}

#########################################################################

function ftp_close
{
    (( _ftp_debug > 0 )) && print -u2 "DEBUG: $0 $@"
    (( _ftp_debug > 3 )) && set -$_ftp_kshflags

    ((_ftp_pi_pid  > 0 )) && _ftp_shutdown_pi
    ((_ftp_dtp_pid > 0 )) && _ftp_terminate_process _ftp_dtp_pid
    return 0
}

#########################################################################

function ftp_login {
    (( _ftp_debug > 0 )) && print -u2 "DEBUG: $0 $@"
    (( _ftp_debug > 3 )) && set -$_ftp_kshflags

    (( $# < 1 )) && _ftp_fatal "$0: name [passwd] [account]"
    typeset user=$1 pass=$2 acct=$3

    if _ftp_sendline USER "$user" && _ftp_getack
    then
	if  [[ -n $pass ]] &&
		_ftp_sendline PASS "$pass" && _ftp_getack &&
	    [[ -n $acct ]] &&
		_ftp_sendline ACCT "$acct" && _ftp_getack
	then
	    print -u2 "$0: DEBUG: login successful"
	    return 0
	fi
    fi
}

#########################################################################

function ftp_dir {
    (( _ftp_debug > 0 )) && print -u2 "DEBUG: $0 $@"
    (( _ftp_debug > 3 )) && set -$_ftp_kshflags

    typeset output

    while getopts o: opt
    do
    	case "$opt" in
	    o)	output=$OPTARG;;
	    *)	_ftp_fatal "$0: invalid argument: $opt";;
	esac
    done
    shift OPTIND-1

    # Start data transfer process ("dtp")

    if [[ -n $output ]]
    then _ftp_setup_dtp -p _ftp_dtp_pid > "$output"
    else _ftp_setup_dtp -p _ftp_dtp_pid
    fi

    if _ftp_sendline LIST "$@" && _ftp_getack
    then

	# Ignore status message like "150 Opening ... connection",
	# wait for positive or negative completion message

	while [[ $ftp_code1 -lt 2 ]]
	do _ftp_getack
	done
	print -u2 "DEBUG: LIST completed"

	# The control connection told us the data transfer is
	# finished. Now wait for the dtp to terminate.

	wait $_ftp_dtp_pid
    else
    	print -u2 "DEBUG: LIST failed: $ftp_response"
	return 1
    fi
    return 0
}

#########################################################################

function ftp_get # remfile [locfile]
{
    (( _ftp_debug > 0 )) && print -u2 "DEBUG: $0 $@"
    (( _ftp_debug > 3 )) && set -$_ftp_kshflags
    (( $# == 1 || $# == 2 )) || _ftp_fatal "$0 filename [local name]"

    typeset remname=$1
    typeset locname=${2:-$1}

    _ftp_setup_dtp -p _ftp_dtp_pid -o "-w 5 -v" > "$locname" #& _ftp_dtp_pid=$!
    if _ftp_sendline RETR "$remname" && _ftp_getack
    then
	while [[ $ftp_code1 -lt 2 ]]
	do _ftp_getack
	done
	print -u2 "DEBUG: RETR completed"
	wait $_ftp_dtp_pid
	return $?
    fi
    return 1
}

########################################################################

function ftp_put # locfile [remfile]
{
    (( _ftp_debug > 0 )) && print -u2 "DEBUG: $0 $@"
    (( _ftp_debug > 3 )) && set -$_ftp_kshflags
    (( $# == 1 || $# == 2 )) || _ftp_fatal "$0: local_name [remote_name]"

    typeset locname=$1
    typeset remname=${2:-$1}

    # The dtp should close the TCP connection immediately after
    # transferring all the file data to the server. Unfortunatly
    # "netcat" does not close the connection this way, because it
    # indefinitly waits for the server to initiate a
    # disconnection.
    # This is the reason why we set a timeout of one second ("-w
    # 1") for the connection, after all data was transferred
    # successfully.

    _ftp_setup_dtp -o "-w 1 -v" -p _ftp_dtp_pid < "$locname" #& _ftp_dtp_pid=$!
    sleep 1				# Give the server time to start up
    if _ftp_sendline STOR "$remname" && _ftp_getack
    then
	while [[ $ftp_code1 -lt 2 ]]
	do _ftp_getack
	done
	print -u2 "STOR completed"
	wait $_ftp_dtp_pid
    fi
    return $?
}

########################################################################

function ftp_cd # directory
{
    (( _ftp_debug > 0 )) && print -u2 "DEBUG: $0 $@"
    (( _ftp_debug > 3 )) && set -$_ftp_kshflags
    (( $# < 0 || $# > 1 )) && _ftp_fatal "$0 [directory]"

    typeset dir=$1

    if [[ -n $dir ]]
    then _ftp_sendline CWD "$dir"
    else _ftp_sendline CDUP
    fi
    _ftp_getack
    return $?
}

########################################################################

function ftp_pwd
{
    (( _ftp_debug > 0 )) && print -u2 "DEBUG: $0 $@"
    (( _ftp_debug > 3 )) && set -$_ftp_kshflags
    (( $# != 1 )) && _ftp_fatal "$0 [pwd]"

    typeset pwdvname=$1
    typeset cwd

    if _ftp_sendline PWD && _ftp_getack
    then
    	# example: "/incoming" is current directory.
    	cwd=${ftp_response#\"}
	cwd=${cwd%\"*}
	[[ $pwdvname = - ]] && print "$cwd" || eval $pwdvname='"$cwd"'
    fi
    return $?
}

########################################################################

function ftp_passive
{
    (( _ftp_debug > 0 )) && print -u2 "DEBUG: $0 $@"
    (( _ftp_debug > 3 )) && set -$_ftp_kshflags

    _ftp_passive=$((!_ftp_passive))
}

########################################################################

function ftp_binary
{
    (( _ftp_debug > 0 )) && print -u2 "DEBUG: $0 $@"
    (( _ftp_debug > 3 )) && set -$_ftp_kshflags

    _ftp_sendline "TYPE I" && _ftp_getack
}

########################################################################

function ftp_ascii
{
    (( _ftp_debug > 0 )) && print -u2 "DEBUG: $0 $@"
    (( _ftp_debug > 3 )) && set -$_ftp_kshflags

    _ftp_sendline "TYPE A" && _ftp_getack
}

#########################################################################

function ftp_verbose
{
    (( _ftp_debug > 0 )) && print -u2 "DEBUG: $0 $@"
    (( _ftp_debug > 3 )) && set -$_ftp_kshflags

    _ftp_verbose=$((!_ftp_verbose))
}

function ftp_debug
{
    (( _ftp_debug > 0 )) && print -u2 "DEBUG: $0 $@"
    (( _ftp_debug > 3 )) && set -$_ftp_kshflags

    _ftp_debug=${1:-1}
}

#########################################################################
#########################################################################
# function ftp_authorize # [auth [resp]]
# function ftp_site # args
# function ftp_type # type [args]
# function ftp_ascii, ftp_binary, ftp_ebcdic, ftp_byte
# function ftp_rename # oldname newname
# function ftp_delete # filename
# function ftp_cwd # [dir]
# function ftp_cdup
# function ftp_pwd
# function ftp_rmdir # dir
# function ftp_mkdir # [-r] dir
# function ftp_ls # [dir]
# function ftp_dir # [dir]
# function ftp_get # remfile [locfile [where]]
# function ftp_put # locfile [remfile]
# function ftp_put_unique # locfile [remfile]
# function ftp_append # locfile [remfile]
# function ftp_unique_name # name
# function ftp_mdtm # file
# function ftp_size # file
# function ftp_supported # cmd
# function ftp_nlst # [dir]
# function ftp_list # [dir]
# function ftp_retr # file
# function ftp_stor # file
# function ftp_stou # file
# function ftp_appe # file
# function ftp_port # [port]
# function ftp_pasv
# function ftp_abort
# function ftp_quit
#########################################################################

set +vx
: '
# FTP COMMANDS (RFC 959)

# Access control commands
USER NAME (USER)
PASSWORD (PASS)
ACCOUNT (ACCT)
CHANGE WORKING DIRECTORY (CWD)
CHANGE TO PARENT DIRECTORY (CDUP)
STRUCTURE MOUNT (SMNT)
REINITIALIZE (REIN)
LOGOUT (QUIT)

# Transfer parameter commands
DATA PORT (PORT)
PASSIVE (PASV)
REPRESENTATION TYPE (TYPE)
FILE STRUCTURE (STRU)
TRANSFER MODE (MODE)

# FTP service commands
RETRIEVE (RETR)
STORE (STOR)
STORE UNIQUE (STOU)
APPEND (with create) (APPE)
ALLOCATE (ALLO)
RESTART (REST)
RENAME FROM (RNFR)
RENAME TO (RNTO)
ABORT (ABOR)
DELETE (DELE)
REMOVE DIRECTORY (RMD)
MAKE DIRECTORY (MKD)
PRINT WORKING DIRECTORY (PWD)
LIST (LIST)
NAME LIST (NLST)
SITE PARAMETERS (SITE)
SYSTEM (SYST)
STATUS (STAT)
HELP (HELP)
NOOP (NOOP)

# FTP reply codes
1yz   Positive Preliminary reply
2yz   Positive Completion reply
3yz   Positive Intermediate reply
4yz   Transient Negative Completion reply
5yz   Permanent Negative Completion reply

x0z   Syntax - These replies refer to syntax errors,
      syntactically correct commands that do not fit any functional
      category, unimplemented or superfluous commands.

x1z   Information -  These are replies to requests for
      information, such as status or help.

x2z   Connections - Replies referring to the control and data
      connections.

x3z   Authentication and accounting - Replies for the login
      process and accounting procedures.

x4z   Unspecified as yet.

x5z   File system - These replies indicate the status of the
      Server file system vis-a-vis the requested transfer or other
      file system action.
'
